/*Chris Busch
net-life (c) 1994  
 */

#include <stdlib.h>
#include <iostream.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <signal.h>

#include "netmain.h"

#include "screenio.h"

#include "prob.h"

char* Title::get()
{
  return "<< net-life v1.0.0 by Chris Busch (c) 1994 >>"  ;
}
char* Title::date()
{
  return "( " __DATE__ " " __TIME__ " )";
}	
Title title;

#include "world.h"

#include "debughelp.h"


world earth;
int minalive=50;
int automatic=0;
int visual=0;
int maxturns=3000;
debughelpt debugger;


void mysignalhandler(int sig);

int openlogfile(char *name);
#include "netmenu.cc"


void mysignalhandler(int sig)
{
  int wasactive=sio.isactive();
  if(wasactive){
	sio.uninitkey();
	sio.uninit();
  }
  cout << "\nThis signal was caught #" << sig << "\n";
  tlog << "\nThis signal was caught #" << sig << "\n" << flush;
  switch(sig)
    {
    case 2: //SIGINT:
      if(yesorno("Do you wish to exit so harshly? "))
		{
		  exit(-sig);
		}
    case 16:  //SIGUSR1:
      automatic= (automatic <= 0 ) ? automatic : 1;
      cout << "Resetting automatic rounds to " << automatic << "\n";
      break;
    case 11:  //SIGSEGV:
      cout << "A segmentation fault occured!";
    default:
      cout << "This signal error might be harmful...\n";
      if(yesorno(" Maybe you should exit? [y/n]"))
		{
		  exit(-sig);
		}
    }
  cout << flush;
  if(wasactive)
    {
      sio.init();
	  sio.initkey();
      sio.repaint();
    }
}

ofstream tlog;

int openlogfile(char *name)
{
  char buf[256]="";
  time_t thetime=time(NULL);
  if(name==NULL || name[0]=='\0' )
	//sprintf(buf,"net-life_%d.log",thetime);
	strftime(buf,149,"net-life-%I_%M_%S_%p-%b_%d.log", 
		 localtime( &thetime));
  else
	strcpy(buf,name);

  tlog.open(buf,ios::out | ios::trunc);
  if(!tlog.good())
	{
	  cerr << "could not open log file '" 
		<< buf << "' , will try /tmp dir\n";
	  if(name==NULL)
		sprintf(buf,"/tmp/%s_net-life_%d.log",
				cuserid(NULL),time(NULL));
	  else
		sprintf(buf,"/tmp/%s_%s",cuserid(NULL),name);
	  tlog.open(buf,ios::out);
	}
  if(! tlog.good())
	{
	  cerr << "Could not open log file '" << buf << "'.\n";
	  return 0;
	}
  tlog << title.get() << "\n";
  tlog << "time: " << time(NULL) << "\n";
  cout << "Net-life log file name is '" << buf << "'.\n";
  return 1;
}

int dobatch=0;
char batchfile[80]="";

int verbosebatch=0;

void main(int argc,char **argv)
{
  cout << "\n";
  debugger.settrap(mysignalhandler);
  cout.sync_with_stdio();
  cout << "\n";

  char logfileopened=0,ask2save=1;
  opterr = 0; /* shut off error */
  int c;
  while ((c = getopt(argc, argv, "f:b:vl:ahA")) != EOF)
	switch(c) 
	  {
	  case 'f':
	  case 'b':
		dobatch=1;
		strcpy(batchfile,optarg);
		break;
	  case 'v':
		verbosebatch=1;
		break;
	  case 'l':
		if(!openlogfile(optarg))
		  exit(1);
		logfileopened=1;
		break;
	  case 'A':
		ask2save=0;
		break;
	  case 'a':
	  case 'h':
	  default:
		cout << "\nNet-life is a simulation of A-Life,\n"
		  "with neural nets generated via evolutionary techniques.\n"
		  "(artificial neural nets and evolutionary algorithms)\n" 
		  "\nCommand line args:\n"
		  "-A        ask to save simulation when quitting\n"
		  "-b file   run file as a batch\n"
		  "-f file   see -b\n"
		  "-v        turn on verbose batch\n"
          "-l file   set logfile to file\n"        
		  "\npress h for menu help.\n\n";

	  }

  if(!logfileopened && !openlogfile(optarg))
	exit(1);

  if(dobatch){
	cout << "running batch...\n";
	runbatch(batchfile);
	exit(1);
  }

  visual=yesorno("Run visually? [y/n]");
  

  runany();

  if(ask2save && yesorno("Save it? ")) saveit();

  fflush(stdout);
  cout << flush;
}






