/*Chris Busch
 */

#include <string.h>
#include <ctype.h>
#include <fstream.h>
#include "prob.h"
#include "world.h"

world::world()
{
  starttime=0;
  maxrow=maxcol=0;
  visual=true;
  trail=traildef;
  round=0;
  mapdata=NULL;
  mutnewchance=1.0;
  mutnewamount=10.0;
  mutdupchance=0.07;
  mutdupamount=0.5;
  suppressamount=6.0;
  resizescreen(24,80);
  clear(); 
  init();
}

int world::clear()
{
  turns=0;
  foodworth=foodworthdef;
  for(int r=0;r<maxrow;r++)
    for(int c=0;c<maxcol;c++)
      {
        map(r,c)=0;
      }
  
  return maxrow*maxcol;
}


int world::resizescreen(int row,int col) 
{
  char* newdata=NULL;
  
  if( row!=maxrow || col != maxcol ) 
    { 

      if( (newdata = new char [row*col]) == NULL)
        {
          cerr << "\nlow on memory, cannot resize screen\n";
          return 0;
        }

      memset(newdata,0,sizeof(char)*row*col);
      
      //relocate old map info
      for(int r=0; r<maxrow && r<row ; r++)
        for(int c=0; c<maxcol && c< col;c++)
          {
            newdata[r*col+c]=mapdata[r*maxcol+c];
          }

      if(mapdata != NULL) delete [] mapdata;
      mapdata=newdata;
      maxrow=row; maxcol=col;
      return 1; 
    } 
  return 0; 
} 

void world::draw()
{
  if(visual)
    {
      if(sio.changed())
        {
          fflush(stdin);
          sio.uninitkey();
          sio.uninit();
          sio.init();
          sio.initkey();
          resizescreen();
          sio.cls();

          sio.repaint(); 
          sio.clearchanged();
        }
      if(maxcol < strlen(title.get()))
        sio.puttext(sio.xsize()/2 - 6,0,"<<net-life>>");
      else
        sio.puttext(sio.xsize()/2 - strlen(title.get())/2,0,title.get());

      for(int r=0;r<maxrow;r++)
        for(int c=0;c<maxcol;c++)
          {
            if(sio.changed()) break;
            if(map(r,c)>=begdude && map(r,c)<=enddude+1)
              sio.putpixel(c,r+1,int( convert(map(r,c)) | A_STANDOUT 
#ifdef USECOLORS
         | (COLOR_PAIR((color_pairs[COLOR_YELLOW]))) | A_BOLD
#endif
                                     ));
            else if(map(r,c)>=begdude+dudecoloroffset1 
                    && map(r,c)<=enddude+1+dudecoloroffset1 )
              sio.putpixel(
                           c,r+1,int((convert(map(r,c) ) - dudecoloroffset1)
                                     | A_UNDERLINE | A_REVERSE
#ifdef USECOLORS
     | (COLOR_PAIR(color_pairs[random(COLOR_PAIRS)])) | A_BOLD
#endif
                                     ));
            else
              sio.putpixel(c,r+1,int( convert(map(r,c)) | A_DIM ));
          }

      //draw bottom information bar
      char buf[100];
      int diff=int(time(NULL)-starttime);
      if(maxcol < 60)
        sprintf(buf
                ,"   Round: %d Turn: %d Alive: %d   "
                ,round,turns,numalive);
      else if(diff)
        sprintf(buf
              ,"    Round: %d Turn: %d Alive: %d Map(r%d,c%d) Turns/sec=%d   "
                ,round,turns,numalive,maxrow,maxcol,turns/diff);
      else
        sprintf(buf
                ,"    Round: %d Turn: %d Alive: %d Map(r%d,c%d)   "
                ,round,turns,numalive,maxrow,maxcol);           
      sio.puttext(sio.xsize()/2 - strlen(buf)/2,sio.ysize()-1,buf);
      sio.refresh();
    }
}

void world::init()
{        
  for(int i=0;i<maxdudes;i++)
    {
      dudes[i].x=random()%maxcol;
      dudes[i].y=random()%maxrow;
      dudes[i].mutate();
    }
}

void world::repop(int minfood)
{        
  clear();
  int count=0; //count of alive
  for(int i=0;i<maxdudes;i++)
    if(dudes[i].food > minfood) count++;

  if(count==maxdudes) return;  //none dead!

  int numofkids=(maxdudes-count)/2;

  int lastdead=0;

  if(count>0)  //some are still alive so they can now breed
    {

      for( i=0;i<maxdudes;i++)
        if(dudes[i].food <= minfood) 
          {  //find the first none dead dude
            lastdead=i;
            break;
          }

      //breed numofkids kids
      
      while(numofkids>0)
        for(i=0;numofkids>0 && i < maxdudes; i++)
          {
            if(dudes[i].food > minfood+1) 
              {
                dudes[lastdead].brain.copy(dudes[i].brain);
                dudes[lastdead].food=minfood+1; //clever but HACKISH
                dudes[lastdead].mutate(mutdupchance,mutdupamount);
                dudes[lastdead].age=dudes[i].age/3;
                for(int a=lastdead+1;a<maxdudes;a++)
                  if(dudes[a].food <= minfood) 
                    {
                      lastdead=a;           
                      a=maxdudes+1; //quit loop, nervous about break;
                    }
                numofkids--;
              }
          } 
    }

  //now make alive the rest of the dead ones

  for(i=0;i<maxdudes;i++)  //make a brand new batch!
    {
      if(dudes[i].food <= minfood)
        { //dead ones are being reborn
          dudes[i].mutate(mutnewchance,mutnewamount);
          dudes[i].age=0;
          if(i & 1) dudes[i].suppress(suppressamount);  //suppress only odds
        }
      dudes[i].food=dudes[i].initialfood; //everyone gets this renewed...
      dudes[i].x=random()%maxcol; 
      dudes[i].y=random()%maxrow;
    }
}


int world::isfood(char c)
{
  return (c>=begfood && c<=endfood) ? 1 : 0;
}

int world::foodstrength(int y,int x)
{
  if(y<0 || y>=maxrow) return 0;
  if(x<0 || x>=maxcol) return 0;
  if(isfood(look(y,x)))
    return int(((look(y,x)-begfood+1.0)/foodrange)*foodworth+0.0);
  return 0;
}

prob world::foodpower(int y,int x)
{
  return sigmoid( ((
                    foodstrength(y,x)*3+
                    foodstrength(y-1,x)+
                    foodstrength(y+1,x)+
                    foodstrength(y,x-1)+
                    foodstrength(y,x+1))/6.0)/(1.0+foodworth/12.0),-5); 
  //with the -6 sigma the domain is about 0-10
}

int world::wraprow(int y)
{
  return abs((y<0 ? maxrow+y:y)%maxrow);
}
int world::wrapcol(int y)
{
  return abs((y<0 ? maxcol+y:y)%maxcol);
}

void world::nextmove(critter& c)
{
  float dx,dy;
  dy=c.outup*-2+c.outdn*2;
  dx=c.outlt*-2+c.outrt*2;
  
  c.y+=(int)dy;
  c.x+=(int)dx;

  c.y= (c.y < 0 ) ? 0 : (c.y>=maxrow) ? maxrow-1: c.y;
  c.x= (c.x < 0 ) ? 0 : (c.x>=maxcol) ? maxcol-1: c.x;
  
  
  //    c.y=wraprow(c.y);
  //    c.x=wrapcol(c.x);
}

int world::doeat(critter& c) //assumed after newmove !!
{  
  return c.food+=int(foodstrength(c.y,c.x)*c.outeat);
}


int world::imagefunc(critter& c)
{
  //  return begdude+(int)(duderange*sigmoid(c.food/(c.initialfood/5),-5));
  return int(begdude+duderange*linear(c.food,c.initialfood/2,c.initialfood)
    + ((c.outeat*2) > 1) * dudecoloroffset1);
}


int world::domoves()
{
  int image;
  int numalive=0;
  for(int i=0;i<maxdudes;i++)
    {
      if(!dudes[i].nofood())
        {
          numalive++;
          dudes[i].age++;
          look(dudes[i].y,dudes[i].x)=trail;
          dudes[i].senses(
                          foodpower(dudes[i].y-1,dudes[i].x), //up
                          foodpower(dudes[i].y+1,dudes[i].x), //down
                          foodpower(dudes[i].y,dudes[i].x-1), //left
                          foodpower(dudes[i].y,dudes[i].x+1), //right
                          (dudes[i].age%11)/10.0);
          //                      random.percent()); //random
          //                  1-linear(dudes[i].food,250,500));                 //hunger
          dudes[i].brain.run();
          dudes[i].react();
          dudes[i].modfood();

          nextmove(dudes[i]);
          doeat(dudes[i]);
          //cerr << "now food is::::" << dudes[i].food << "\n";
          if(dudes[i].nofood())
            image=death;
          else
            image=imagefunc(dudes[i]);
          look(dudes[i].y,dudes[i].x)=image;
        }
    }
  return numalive;
}

int world::sprinklefood(int howmany,int howmuch,prob chance)
{
  howmuch= (howmuch>(endfood-begfood)) ? (endfood-begfood) : howmuch;
  for(int i=0;i<howmany;i++)
    {
      if(random.percent()<=chance)
        map(random()%maxrow,random()%maxcol)
          =convert(begfood+(random()%howmuch));
    }
  return howmany;
}

int world::fillwithfood(int howmuch,prob chance)
{
  if(howmuch!=-1)
    howmuch= (howmuch>(endfood-begfood)) ? (endfood-begfood) : howmuch;
  for(int r=0;r<maxrow;r++)
    for(int c=0;c<maxcol;c++)
      {
        if(random.percent()<=chance)
          {
            if(howmuch==-1)
              look(r,c)=convert(blank);
            else
              look(r,c)=convert(begfood+(random()%howmuch));
          }
      }
  return howmuch;
}

int world::runwithfood(int howmany,int howmuch,prob chance)
{
  static int xx=random(maxcol),yy=random(maxrow),tx=xx,ty=yy;

  howmuch= (howmuch> foodrange) ? foodrange : howmuch;

  for(int i=0;i<howmany;i++)
    if(random.percent()<=chance)
      {
        while(xx==tx && yy==ty)
          {
            tx=random(maxcol);
            ty=random(maxrow);
          }

        yy+= random(2)-(yy>ty);
        xx+= random(2)-(xx>tx);

        look(yy,xx)=convert(begfood+random(howmuch));
      }

  return howmany;
}


int world::runturn()
{
  numalive=domoves();
  draw();
  turns++;
  return numalive;
}


void world::dump(int foodlevel)
{
  cout << "world::[" << this << "] turns=" << turns << "\n";
  for(int i=0;i<maxdudes;i++)
    {
      if(foodlevel==-1 || dudes[i].food>foodlevel )
        {
          cout << "dude[" << i << "]\n";
          dudes[i].dump();
        }
    }
}


static char* tounderscore(char s[])
{
  for(int i=strlen(s)-1;i>=0;i--)
    if(!isalnum(s[i])) s[i]='_';
  return s;
}

cgbbool world::save(char fname[],char title[])
{
  ofstream out(fname);
  if(out.bad())
    {
      cerr << "cannot open filename: " << fname << "\n";
      return false;
    }
  out << "\nbegin world " << tounderscore(title) << "\n";
  out << "Version " << Version << "\n";
  out << "round " << round << "\n";
  out << "fillwithfood " << fwfparam.quality << " " 
    << fwfparam.quantity << " " << fwfparam.chance << "\n";
  out << "runwithfood " << rwfparam.quality << " " 
    << rwfparam.quantity << " " << rwfparam.chance << "\n";
  out << "sprinklefood " << sfparam.quality << " " 
    << sfparam.quantity << " " << sfparam.chance << "\n";

  for(int i=0;i<maxdudes;i++)
    {
      out << "\nbegin critter " << i << "\n";
      if(!dudes[i].save(out))
        return false;
      out << "end critter\n";
    }
  out << "\nend world\n";
  return true;
}

cgbbool world::load(char fname[])
{
  ifstream filew(fname);
  if( !filew || filew.bad())
    {
      cerr << "cannot open filename: " << fname << "\n";
      return false;
    }
  char dummy[100];
  //filew << "\nbegin world " << tounderscore(title) << "\n";
  filew >> dummy >> dummy >> dummy;
  cout << "reading in " << dummy << ".\n";
  int fileVersion;
  //out << "Version " << Version;
  filew >> dummy >> fileVersion;
  if(Version != fileVersion)
    {
      cerr << "Wrong Version!  Current version (v" 
        << Version << ") is not equal to that version (v" 
          << fileVersion << ").\n";
      return false;
    }

  //out << "round " << round;
  filew >> dummy >> round;

  //  out << "fillwithfood " << fwfparam.quality << " " 
  //    << fwfparam.quantity << " " << fwfparam.chance << "\n";
  filew >> dummy >>  fwfparam.quality >> fwfparam.quantity >> fwfparam.chance;
  //  out << "runwithfood " << rwfparam.quality << " " 
  //    << rwfparam.quantity << " " << rwfparam.chance << "\n";
  filew >> dummy >>  rwfparam.quality >> rwfparam.quantity >> rwfparam.chance;
  //  out << "sprinklefood " << sfparam.quality << " " 
  //    << sfparam.quantity << " " << sfparam.chance << "\n";
  filew >> dummy >>  sfparam.quality >> sfparam.quantity >> sfparam.chance;

  for(int i=0;i<maxdudes;i++)
    {
      //filew << "\nbegin critter " << i << "\n";
      filew >> dummy >> dummy >> dummy;
      if(!dudes[i].load(filew))
        return false;
      //filew << "end critter\n";
      filew >> dummy >> dummy;
    }
  //filew << "\nend world\n";
  filew >> dummy >> dummy;
  if(filew.bad()) return false;
  return true;
}





//end of this file




